@@ -33,6 +33,7 @@ import (
 	"github.com/prometheus/client_golang/api"
 	v1 "github.com/prometheus/client_golang/api/prometheus/v1"
 	"github.com/prometheus/client_golang/prometheus/promhttp"
+	"github.com/prometheus/client_golang/prometheus/testutil/promlint"
 	config_util "github.com/prometheus/common/config"
 	"github.com/prometheus/common/model"
 	"github.com/prometheus/common/version"
@@ -40,7 +41,6 @@ import (
 
 	"github.com/prometheus/prometheus/config"
 	"github.com/prometheus/prometheus/pkg/rulefmt"
-	"github.com/prometheus/prometheus/util/promlint"
 )
 
 func main() {
@@ -95,6 +95,8 @@ func main() {
 	queryLabelsCmd := queryCmd.Command("labels", "Run labels query.")
 	queryLabelsServer := queryLabelsCmd.Arg("server", "Prometheus server to query.").Required().URL()
 	queryLabelsName := queryLabelsCmd.Arg("name", "Label name to provide label values for.").Required().String()
+	queryLabelsBegin := queryLabelsCmd.Flag("start", "Start time (RFC3339 or Unix timestamp).").String()
+	queryLabelsEnd := queryLabelsCmd.Flag("end", "End time (RFC3339 or Unix timestamp).").String()
 
 	testCmd := app.Command("test", "Unit testing.")
 	testRulesCmd := testCmd.Command("rules", "Unit tests for rules.")
@@ -142,7 +144,7 @@ func main() {
 		os.Exit(debugAll(*debugAllServer))
 
 	case queryLabelsCmd.FullCommand():
-		os.Exit(QueryLabels(*queryLabelsServer, *queryLabelsName, p))
+		os.Exit(QueryLabels(*queryLabelsServer, *queryLabelsName, *queryLabelsBegin, *queryLabelsEnd, p))
 
 	case testRulesCmd.FullCommand():
 		os.Exit(RulesUnitTest(*testRulesFiles...))
@@ -324,19 +326,19 @@ type compareRuleType struct {
 	label  map[string]string
 }
 
-func checkDuplicates(r []rulefmt.RuleGroup) []compareRuleType {
+func checkDuplicates(groups []rulefmt.RuleGroup) []compareRuleType {
 	var duplicates []compareRuleType
 
-	for rindex := range r {
-		for index, props := range r[rindex].Rules {
+	for _, group := range groups {
+		for index, rule := range group.Rules {
 			inst := compareRuleType{
-				metric: props.Record,
-				label:  props.Labels,
+				metric: ruleMetric(rule),
+				label:  rule.Labels,
 			}
 			for i := 0; i < index; i++ {
 				t := compareRuleType{
-					metric: r[rindex].Rules[i].Record,
-					label:  r[rindex].Rules[i].Labels,
+					metric: ruleMetric(group.Rules[i]),
+					label:  group.Rules[i].Labels,
 				}
 				if reflect.DeepEqual(t, inst) {
 					duplicates = append(duplicates, t)
@@ -347,6 +349,13 @@ func checkDuplicates(r []rulefmt.RuleGroup) []compareRuleType {
 	return duplicates
 }
 
+func ruleMetric(rule rulefmt.RuleNode) string {
+	if rule.Alert.Value != "" {
+		return rule.Alert.Value
+	}
+	return rule.Record.Value
+}
+
 var checkMetricsUsage = strings.TrimSpace(`
 Pass Prometheus metrics over stdin to lint them for consistency and correctness.
 
@@ -488,30 +497,10 @@ func QuerySeries(url *url.URL, matchers []string, start, end string, p printer)
 		return 1
 	}
 
-	// TODO: clean up timestamps
-	var (
-		minTime = time.Now().Add(-9999 * time.Hour)
-		maxTime = time.Now().Add(9999 * time.Hour)
-	)
-
-	var stime, etime time.Time
-
-	if start == "" {
-		stime = minTime
-	} else {
-		stime, err = parseTime(start)
-		if err != nil {
-			fmt.Fprintln(os.Stderr, "error parsing start time:", err)
-		}
-	}
-
-	if end == "" {
-		etime = maxTime
-	} else {
-		etime, err = parseTime(end)
-		if err != nil {
-			fmt.Fprintln(os.Stderr, "error parsing end time:", err)
-		}
+	stime, etime, err := parseStartTimeAndEndTime(start, end)
+	if err != nil {
+		fmt.Fprintln(os.Stderr, err)
+		return 1
 	}
 
 	// Run query against client.
@@ -530,7 +519,7 @@ func QuerySeries(url *url.URL, matchers []string, start, end string, p printer)
 }
 
 // QueryLabels queries for label values against a Prometheus server.
-func QueryLabels(url *url.URL, name string, p printer) int {
+func QueryLabels(url *url.URL, name string, start, end string, p printer) int {
 	config := api.Config{
 		Address: url.String(),
 	}
@@ -542,10 +531,16 @@ func QueryLabels(url *url.URL, name string, p printer) int {
 		return 1
 	}
 
+	stime, etime, err := parseStartTimeAndEndTime(start, end)
+	if err != nil {
+		fmt.Fprintln(os.Stderr, err)
+		return 1
+	}
+
 	// Run query against client.
 	api := v1.NewAPI(c)
 	ctx, cancel := context.WithTimeout(context.Background(), 2*time.Minute)
-	val, warn, err := api.LabelValues(ctx, name)
+	val, warn, err := api.LabelValues(ctx, name, stime, etime)
 	cancel()
 
 	for _, v := range warn {
@@ -561,10 +556,36 @@ func QueryLabels(url *url.URL, name string, p printer) int {
 	return 0
 }
 
+func parseStartTimeAndEndTime(start, end string) (time.Time, time.Time, error) {
+	var (
+		minTime = time.Now().Add(-9999 * time.Hour)
+		maxTime = time.Now().Add(9999 * time.Hour)
+		err     error
+	)
+
+	stime := minTime
+	etime := maxTime
+
+	if start != "" {
+		stime, err = parseTime(start)
+		if err != nil {
+			return stime, etime, errors.Wrap(err, "error parsing start time")
+		}
+	}
+
+	if end != "" {
+		etime, err = parseTime(end)
+		if err != nil {
+			return stime, etime, errors.Wrap(err, "error parsing end time")
+		}
+	}
+	return stime, etime, nil
+}
+
 func parseTime(s string) (time.Time, error) {
 	if t, err := strconv.ParseFloat(s, 64); err == nil {
 		s, ns := math.Modf(t)
-		return time.Unix(int64(s), int64(ns*float64(time.Second))), nil
+		return time.Unix(int64(s), int64(ns*float64(time.Second))).UTC(), nil
 	}
 	if t, err := time.Parse(time.RFC3339Nano, s); err == nil {
 		return t, nil
